"""Module with persistent Img class."""

import importlib
from pathlib import Path
import re
from typing import Tuple, List, Optional

from astropy.io import fits

from micado.main.BaseFrame import BaseFrame
import micado.recipes
from micado.main.DataItem import DataItem
from micado.main.DataItemExtra import KEY_SCI, KEY_ERR, KEY_DQ


class ImgExtra(BaseFrame):
    """An Img DataItem, a processed image."""

    @classmethod
    def layer_names(cls) -> Tuple[List[str], List[str]]:
        """The names of the layers of this data product."""
        # By default no name
        return [KEY_SCI, KEY_ERR, KEY_DQ], []

    @classmethod
    def layers_number(cls) -> Tuple[Optional[int], Optional[int]]:
        """Minimum/maximum number of detectors for each layer."""
        if "HCI" in cls.__name__:
            return 1, 1
        elif "persist" in cls.__name__.lower():
            return 1, 9
        else:
            return 9, 9

    def real_fits_keywords(self, prefixes=None):
        """Get the FITS keywords, overloaded to provide PRO.CATG.

        TODO: Move this to a class higher in the hierarchy so processed
        catalogs will also have a PRO.CATG.
        """
        kws = super().real_fits_keywords(prefixes=prefixes)
        if prefixes is None:
            kws["PRO CATG"] = self.name_oca()
        return kws

    def make(self):
        """Make this object either through prototype code or CPL Recipe."""
        if self.get_cpl_recipe():
            self.make_cpl()
        else:
            self.make_proto()

    def make_proto(self):
        """Make this object.

        TODO: Move this to the Recipes, and split up in RecipeSteps.
        """
        r = importlib.import_module(
            "{}.{}".format(
                micado.recipes.__name__, self.recipe.__class__.__name__ + "Make"
            )
        )
        r.make(self)

    def get_cpl_recipe(self):
        """Get the CPL recipe if available."""
        # TODO: Move to Recipe class?
        # python-cpl is imported here and not at the toplevel so it is
        # possible to use the rest of MicadoWISE when cpl is not available.
        try:
            import cpl
        except ModuleNotFoundError as e:
            print(f"Warning: cannot import CPL: {e}")
            return

        try:
            cpl.esorex.init()
        except FileNotFoundError as e:
            # E.g. FileNotFoundError: [Errno 2] No such file or directory:
            #      '/home/hugo/.esorex/esorex.rc'
            print(f"Warning: cannot initialize ESOREX: {e}")
            return

        recipes = dict(cpl.Recipe.list())
        if self.recipe.name_eso() not in recipes:
            return None

        return cpl.Recipe(self.recipe.name_eso())

    def make_cpl(self):
        """Make this object through CPL Recipe.

        Requires CPL to be installed.
        """
        # Assert the data is not already made.
        assert self.extensions in [None, []]
        recipe = self.get_cpl_recipe()
        # TODO: set recipe.param
        sof_dict = self.get_sof_dict()
        result = recipe(raw=sof_dict)
        self.result = result
        # TODO: Catch Exceptions?
        hdulist = getattr(result, self.name_oca())

        # TODO: Check whether header keywords are as expected, e.g.
        #       filenames and such.
        # TODO: This requires MicadoWISE to adjust to the ESO scheme first.

        # MAJOR TODO: Copy generated information from the FITS headers.
        # E.g. QC Parameters.

        # Set the pixel values.
        # TODO: Refactor this with _construct_object_from_fitsfile
        hdus: List[fits.ImageHDU] = [
            hdu for hdu in hdulist[1:] if isinstance(hdu, fits.ImageHDU)
        ]
        # "DET3.DATA" -> ('3', 'DATA')
        pattern_extname = re.compile("^DET(\d)\.(\w+)$")
        for hdu in hdus:
            assert "EXTNAME" in hdu.header, "Cannot find EXTNAME"
            extname = hdu.header["EXTNAME"]
            assert re.match(pattern_extname, extname), f"Bad EXTNAME {extname}"
        hdusd = {
            re.match(pattern_extname, hdu.header["EXTNAME"]).groups(): hdu
            for hdu in hdus
        }
        extensionnumbers = sorted(set(k for k, _ in hdusd))
        # E.g. ["1", "2", ..., "9"]
        assert extensionnumbers == ["1", "2", "3", "4", "5", "6", "7", "8", "9"]
        prop = getattr(self.__class__, "extensions")
        extensions = []
        for extensionnr in extensionnumbers:
            hdudata = hdusd[(extensionnr, "DATA")]
            # Here we assume self.extensions are in the correct order.
            extension = prop.prop_type(header=hdudata.header)
            extension.image = hdudata.data
            if (extensionnr, "ERR") in hdusd:
                extension.uncertainty = hdusd[(extensionnr, "ERR")].data
            if (extensionnr, "DQ") in hdusd:
                extension.quality = hdusd[(extensionnr, "DQ")].data
            extensions.append(extension)
        self.extensions = extensions

    def get_sof_dict(self, save=True, usebase=False):
        """Get the SOF dictionary."""
        # Set the input
        sof_dict = {}
        if usebase:
            # Get the base class that shares this recipe.
            thecls = self.__class__.get_base_dataitem()
        else:
            thecls = self.__class__
        props_classes = thecls.get_inputdata_full()
        for prop, dcls in props_classes.items():
            # Get the key for the SOF.
            # TODO: Refactor with save_sof()
            key_sof = dcls.name_oca()

            # Get the actual instances that fill these dependencies.
            insts = getattr(self, prop.attribute)
            if insts is None:
                continue
            elif isinstance(insts, list):
                for instance in insts:
                    if save:
                        instance.save_fits()
                sof_dict[key_sof] = [instance.filename for instance in insts]
            elif isinstance(insts, DataItem):
                if save:
                    insts.save_fits()
                sof_dict[key_sof] = insts.filename
            else:
                raise ValueError(f"Unknown instance {insts}")
        return sof_dict

    def filename_of_sof(self):
        """Filename of the SOF file to create this data product."""
        return Path(self.filename_for_sof()).with_suffix(".sof")

    def save_sof(self, use_name_eso=False, filename=None, usebase=False):
        """Create an SOF file to create this Data Product with ESOREX."""
        # TODO: Move this to a class earlier in the hierarchy so it applies
        #       to catalogs too.
        # TODO: Use get_sof_dict(). get_sof_dict() uses full filename though.
        if filename is None:
            filename = self.filename_of_sof()

        s_sof_lines = []

        # Loop over the inputdata of the class.
        if usebase:
            # Get the base class that shares this recipe.
            thecls = self.__class__.get_base_dataitem()
        else:
            thecls = self.__class__
        props_classes = thecls.get_inputdata_full()
        for prop, dcls in props_classes.items():
            # Get the actual instances that fill these dependencies.
            instances = getattr(self, prop.attribute)
            if not isinstance(instances, list):
                instances = [instances]
            for instance in instances:
                if instance is not None:
                    # Get the filename from the instance, but the CATG from
                    # the class, because that might be a subcategory.
                    if use_name_eso:
                        s_sof_line = (
                            instance.filename_for_sof() + " " + dcls.name_eso() + "\n"
                        )
                    else:
                        s_sof_line = (
                            instance.filename_for_sof() + " " + dcls.name_oca() + "\n"
                        )
                    s_sof_lines.append(s_sof_line)

        s_sof = "".join(s_sof_lines)
        with open(filename, "w") as f:
            f.write(s_sof)
