# -*- coding: utf-8 -*-
"""Tests for Detrend.py."""

from astropy.io import fits

from micado.config.Environment import Env
from micado.hardware.ShotNoise import ShotNoise
from micado.hardware.SummedExposure import SummedExposure
from micado.main.ObservingBlock import ObservingBlock
from micado.hardware.Detector import Detector
from micado.hardware.DetectorArray import DetectorArray
from micado.hardware.DarkCurrent import DarkCurrent
from micado.dataitems.MasterFlat import MasterFlat
from micado.dataitems.Science import Science
from micado.dataitems.ScienceExtension import ScienceExtension
from micado.dataitems.Detrend import Detrend
from micado.dataitems.DepersistedScience import DepersistedScience
from micado.recipes.McdImgDetrend import McdImgDetrend
from micado.templates.MicadoImgObs import MicadoImgObs
from micado.test.functional.science_test import get_filt_inst_ro

from .dark_test import remove_file_and_links
from .masterdark_test import create_masterdark, mock_depersist


def test_processing():
    """Test processing of Detrend."""
    Env["use_python_logging"] = True
    myraws, mymasterdark, dark_real, myraws_raw = create_masterdark()

    # TODO: Do not reuse the MasterDark as MasterFlat.
    # TODO: Create an easy way to generate data products.
    fn_flat = mymasterdark.filename.replace(".fits", "-flat.fits")
    hdus = fits.open(mymasterdark.filename)
    hdus[0].header["MASTERDA"] = None
    hdus.writeto(fn_flat)
    mymasterflat = MasterFlat(fn_flat)

    mydetector = Detector(
        xhw=0.24,
        yhw=0.24,
    )
    mydetectorlist = [mydetector]
    mydarkcurrent = DarkCurrent(value=dark_real)
    myshotnoise = ShotNoise(state=1)

    mydetectorarray = DetectorArray(
        name="testdetector",
        detector_array_list=mydetectorlist,
        dark_current=mydarkcurrent,
        shot_noise=myshotnoise,
        qecurve=None,
        linearity=None,
        exposure_action=SummedExposure(),
    )

    # A template
    science_template = MicadoImgObs(det__dit=1)

    # An observingblock
    science_ob = ObservingBlock(
        name="MyScienceTest",
        instrument="MICADO",
        target="empty_sky",
        templates=[science_template],
    )

    filt1, filt2, inst, ro = get_filt_inst_ro()

    # Extensions
    myextensions = [
        ScienceExtension(detector=det, BITPIX=-64, NAXIS=2) for det in mydetectorlist
    ]

    # Create raw
    myrawscience = Science(
        detector=mydetectorarray,
        instrument=inst,
        relay_optics=ro,
        template=science_template,
        observing_block=science_ob,
        extensions=myextensions,
    )
    myrawscience.make_simulation()
    myrawscience.save_fits()

    mydepersistedscience = mock_depersist(DepersistedScience, myrawscience)
    mydepersistedscience.save_fits()

    myrecipe = McdImgDetrend()
    mydetrend = Detrend(
        raw=mydepersistedscience,
        masterdark=mymasterdark,
        masterflat=mymasterflat,
        recipe=myrecipe,
    )
    mydetrend.make_proto()
    mydetrend.save_fits()

    # TODO: Fix headers
    # verify_fitsheaders(mydetrend.filename)

    # TODO: add tests for pixel values
    # That is, everything should be 0 now

    for dm in (
        myraws
        + myraws_raw
        + [myrawscience, mydepersistedscience, mymasterdark, mymasterflat, mydetrend]
    ):
        remove_file_and_links(dm.filename)

    remove_file_and_links(fn_flat)
